from dataclasses import dataclass, field
from typing import List, Optional

__NAMESPACE__ = "sdformat/v1.4/sensor.xsd"


@dataclass
class Sensor:
    """
    The sensor tag describes the type and properties of a sensor.

    Parameters
    ----------
    always_on: If true the sensor will always be updated according to
        the update rate.
    update_rate: The frequency at which the sensor data is generated. If
        left unspecified, the sensor will generate data every cycle.
    visualize: If true, the sensor is visualized in the GUI
    pose: This is the pose of the sensor, relative to the parent link
        reference frame.
    topic: Name of the topic on which data is published. This is
        necessary for visualization
    plugin: A plugin is a dynamically loaded chunk of code. It can exist
        as a child of world, model, and sensor.
    camera: These elements are specific to camera sensors.
    contact: These elements are specific to the contact sensor.
    force_torque: These elements are specific to the force torque
        sensor.
    gps: These elements are specific to the GPS sensor.
    imu: These elements are specific to the IMU sensor.
    ray: These elements are specific to the ray (laser) sensor.
    rfidtag:
    rfid:
    sonar: These elements are specific to the sonar sensor.
    transceiver: These elements are specific to a wireless transceiver.
    name: A unique name for the sensor. This name must not match another
        model in the model.
    type: The type name of the sensor. By default, SDFormat supports
        types camera, depth, multicamera, contact, gps, imu, ir and ray.
    """

    class Meta:
        name = "sensor"

    always_on: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    update_rate: float = field(
        default=0.0,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    visualize: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    pose: str = field(
        default="0 0 0 0 0 0",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
        },
    )
    topic: str = field(
        default="__default__",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    plugin: List["Sensor.Plugin"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    camera: Optional["Sensor.Camera"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    contact: Optional["Sensor.Contact"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    force_torque: Optional["Sensor.ForceTorque"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    gps: Optional["Sensor.Gps"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    imu: Optional["Sensor.Imu"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    ray: Optional["Sensor.Ray"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    rfidtag: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    rfid: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    sonar: Optional["Sensor.Sonar"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    transceiver: Optional["Sensor.Transceiver"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    type: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class Plugin:
        """A plugin is a dynamically loaded chunk of code.

        It can exist as a child of world, model, and sensor.

        Parameters
        ----------
        any_element: This is a special element that should not be
            specified in an SDFormat file. It automatically copies child
            elements into the SDFormat element so that a plugin can
            access the data.
        name: A unique name for the plugin, scoped to its parent.
        filename: Name of the shared library to load. If the filename is
            not a full path name, the file will be searched for in the
            configuration paths.
        """

        any_element: List[object] = field(
            default_factory=list,
            metadata={
                "type": "Wildcard",
                "namespace": "##any",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )
        filename: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

    @dataclass
    class Camera:
        """
        These elements are specific to camera sensors.

        Parameters
        ----------
        pose: A position and orientation in the parent coordinate frame
            for the camera.
        horizontal_fov: Horizontal field of view
        image: The image size in pixels and format.
        clip: The near and far clip planes. Objects closer or farther
            than these planes are not rendered.
        save: Enable or disable saving of camera frames.
        depth_camera: Depth camera parameters
        noise: The properties of the noise model that should be applied
            to generated images
        name: An optional name for the camera.
        """

        pose: str = field(
            default="0 0 0 0 0 0",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        horizontal_fov: float = field(
            default=1.047,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        image: Optional["Sensor.Camera.Image"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        clip: Optional["Sensor.Camera.Clip"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        save: Optional["Sensor.Camera.Save"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        depth_camera: Optional["Sensor.Camera.DepthCamera"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        noise: Optional["Sensor.Camera.Noise"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        name: str = field(
            default="__default__",
            metadata={
                "type": "Attribute",
            },
        )

        @dataclass
        class Image:
            """
            The image size in pixels and format.

            Parameters
            ----------
            width: Width in pixels
            height: Height in pixels
            format:
                (L8|R8G8B8|B8G8R8|BAYER_RGGB8|BAYER_BGGR8|BAYER_GBRG8|BAYER_GRBG8)
            """

            width: int = field(
                default=320,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            height: int = field(
                default=240,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            format: str = field(
                default="R8G8B8",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Clip:
            """The near and far clip planes.

            Objects closer or farther than these planes are not
            rendered.

            Parameters
            ----------
            near: Near clipping plane
            far: Far clipping plane
            """

            near: float = field(
                default=0.1,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            far: float = field(
                default=100.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Save:
            """
            Enable or disable saving of camera frames.

            Parameters
            ----------
            path: The path name which will hold the frame data. If path
                name is relative, then directory is relative to current
                working directory.
            enabled: True = saving enabled
            """

            path: str = field(
                default="__default__",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            enabled: Optional[bool] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

        @dataclass
        class DepthCamera:
            """
            Depth camera parameters.

            Parameters
            ----------
            output: Type of output
            """

            output: str = field(
                default="depths",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Noise:
            """
            The properties of the noise model that should be applied to
            generated images.

            Parameters
            ----------
            type: The type of noise.  Currently supported types are:
                "gaussian" (draw additive noise values independently for
                each pixel from a Gaussian distribution).
            mean: For type "gaussian," the mean of the Gaussian
                distribution from which noise values are drawn.
            stddev: For type "gaussian," the standard deviation of the
                Gaussian distribution from which noise values are drawn.
            """

            type: str = field(
                default="gaussian",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            mean: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            stddev: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

    @dataclass
    class Contact:
        """
        These elements are specific to the contact sensor.

        Parameters
        ----------
        collision: name of the collision element within a link that acts
            as the contact sensor.
        topic: Topic on which contact data is published.
        """

        collision: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        topic: str = field(
            default="__default_topic__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class ForceTorque:
        """
        These elements are specific to the force torque sensor.

        Parameters
        ----------
        frame: Frame in which to report the wrench values.
        """

        frame: str = field(
            default="parent",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Gps:
        """
        These elements are specific to the GPS sensor.

        Parameters
        ----------
        position_sensing: Parameters related to GPS position
            measurement.
        velocity_sensing: Parameters related to GPS position
            measurement.
        """

        position_sensing: Optional["Sensor.Gps.PositionSensing"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        velocity_sensing: Optional["Sensor.Gps.VelocitySensing"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class PositionSensing:
            """
            Parameters related to GPS position measurement.

            Parameters
            ----------
            horizontal: Noise parameters for horizontal position
                measurement, in units of meters.
            vertical: Noise parameters for vertical position
                measurement, in units of meters.
            """

            horizontal: Optional["Sensor.Gps.PositionSensing.Horizontal"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            vertical: Optional["Sensor.Gps.PositionSensing.Vertical"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

            @dataclass
            class Horizontal:
                """
                Noise parameters for horizontal position measurement, in units
                of meters.

                Parameters
                ----------
                noise: The properties of a sensor noise model.
                """

                noise: Optional["Sensor.Gps.PositionSensing.Horizontal.Noise"] = field(
                    default=None,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

                @dataclass
                class Noise:
                    """
                    The properties of a sensor noise model.

                    Parameters
                    ----------
                    mean: For type "gaussian*", the mean of the Gaussian
                        distribution from which noise values are drawn.
                    stddev: For type "gaussian*", the standard deviation
                        of the Gaussian distribution from which noise
                        values are drawn.
                    bias_mean: For type "gaussian*", the mean of the
                        Gaussian distribution from which bias values are
                        drawn.
                    bias_stddev: For type "gaussian*", the standard
                        deviation of the Gaussian distribution from
                        which bias values are drawn.
                    precision: For type "gaussian_quantized", the
                        precision of output signals. A value       of
                        zero implies infinite precision / no
                        quantization.
                    type: The type of noise. Currently supported types
                        are:       "none" (no noise).       "gaussian"
                        (draw noise values independently for each
                        measurement from a Gaussian distribution).
                        "gaussian_quantized" ("gaussian" plus
                        quantization of outputs (ie. rounding))
                    """

                    mean: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    stddev: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    bias_mean: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    bias_stddev: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    precision: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    type: Optional[str] = field(
                        default=None,
                        metadata={
                            "type": "Attribute",
                            "required": True,
                        },
                    )

            @dataclass
            class Vertical:
                """
                Noise parameters for vertical position measurement, in units of
                meters.

                Parameters
                ----------
                noise: The properties of a sensor noise model.
                """

                noise: Optional["Sensor.Gps.PositionSensing.Vertical.Noise"] = field(
                    default=None,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

                @dataclass
                class Noise:
                    """
                    The properties of a sensor noise model.

                    Parameters
                    ----------
                    mean: For type "gaussian*", the mean of the Gaussian
                        distribution from which noise values are drawn.
                    stddev: For type "gaussian*", the standard deviation
                        of the Gaussian distribution from which noise
                        values are drawn.
                    bias_mean: For type "gaussian*", the mean of the
                        Gaussian distribution from which bias values are
                        drawn.
                    bias_stddev: For type "gaussian*", the standard
                        deviation of the Gaussian distribution from
                        which bias values are drawn.
                    precision: For type "gaussian_quantized", the
                        precision of output signals. A value       of
                        zero implies infinite precision / no
                        quantization.
                    type: The type of noise. Currently supported types
                        are:       "none" (no noise).       "gaussian"
                        (draw noise values independently for each
                        measurement from a Gaussian distribution).
                        "gaussian_quantized" ("gaussian" plus
                        quantization of outputs (ie. rounding))
                    """

                    mean: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    stddev: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    bias_mean: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    bias_stddev: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    precision: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    type: Optional[str] = field(
                        default=None,
                        metadata={
                            "type": "Attribute",
                            "required": True,
                        },
                    )

        @dataclass
        class VelocitySensing:
            """
            Parameters related to GPS position measurement.

            Parameters
            ----------
            horizontal: Noise parameters for horizontal velocity
                measurement, in units of meters/second.
            vertical: Noise parameters for vertical velocity
                measurement, in units of meters/second.
            """

            horizontal: Optional["Sensor.Gps.VelocitySensing.Horizontal"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            vertical: Optional["Sensor.Gps.VelocitySensing.Vertical"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

            @dataclass
            class Horizontal:
                """
                Noise parameters for horizontal velocity measurement, in units
                of meters/second.

                Parameters
                ----------
                noise: The properties of a sensor noise model.
                """

                noise: Optional["Sensor.Gps.VelocitySensing.Horizontal.Noise"] = field(
                    default=None,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

                @dataclass
                class Noise:
                    """
                    The properties of a sensor noise model.

                    Parameters
                    ----------
                    mean: For type "gaussian*", the mean of the Gaussian
                        distribution from which noise values are drawn.
                    stddev: For type "gaussian*", the standard deviation
                        of the Gaussian distribution from which noise
                        values are drawn.
                    bias_mean: For type "gaussian*", the mean of the
                        Gaussian distribution from which bias values are
                        drawn.
                    bias_stddev: For type "gaussian*", the standard
                        deviation of the Gaussian distribution from
                        which bias values are drawn.
                    precision: For type "gaussian_quantized", the
                        precision of output signals. A value       of
                        zero implies infinite precision / no
                        quantization.
                    type: The type of noise. Currently supported types
                        are:       "none" (no noise).       "gaussian"
                        (draw noise values independently for each
                        measurement from a Gaussian distribution).
                        "gaussian_quantized" ("gaussian" plus
                        quantization of outputs (ie. rounding))
                    """

                    mean: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    stddev: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    bias_mean: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    bias_stddev: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    precision: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    type: Optional[str] = field(
                        default=None,
                        metadata={
                            "type": "Attribute",
                            "required": True,
                        },
                    )

            @dataclass
            class Vertical:
                """
                Noise parameters for vertical velocity measurement, in units of
                meters/second.

                Parameters
                ----------
                noise: The properties of a sensor noise model.
                """

                noise: Optional["Sensor.Gps.VelocitySensing.Vertical.Noise"] = field(
                    default=None,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

                @dataclass
                class Noise:
                    """
                    The properties of a sensor noise model.

                    Parameters
                    ----------
                    mean: For type "gaussian*", the mean of the Gaussian
                        distribution from which noise values are drawn.
                    stddev: For type "gaussian*", the standard deviation
                        of the Gaussian distribution from which noise
                        values are drawn.
                    bias_mean: For type "gaussian*", the mean of the
                        Gaussian distribution from which bias values are
                        drawn.
                    bias_stddev: For type "gaussian*", the standard
                        deviation of the Gaussian distribution from
                        which bias values are drawn.
                    precision: For type "gaussian_quantized", the
                        precision of output signals. A value       of
                        zero implies infinite precision / no
                        quantization.
                    type: The type of noise. Currently supported types
                        are:       "none" (no noise).       "gaussian"
                        (draw noise values independently for each
                        measurement from a Gaussian distribution).
                        "gaussian_quantized" ("gaussian" plus
                        quantization of outputs (ie. rounding))
                    """

                    mean: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    stddev: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    bias_mean: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    bias_stddev: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    precision: float = field(
                        default=0.0,
                        metadata={
                            "type": "Element",
                            "namespace": "",
                            "required": True,
                        },
                    )
                    type: Optional[str] = field(
                        default=None,
                        metadata={
                            "type": "Attribute",
                            "required": True,
                        },
                    )

    @dataclass
    class Imu:
        """
        These elements are specific to the IMU sensor.

        Parameters
        ----------
        topic: Topic on which data is published.
        noise: The properties of the noise model that should be applied
            to generated data
        """

        topic: str = field(
            default="__default_topic__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        noise: Optional["Sensor.Imu.Noise"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class Noise:
            """
            The properties of the noise model that should be applied to
            generated data.

            Parameters
            ----------
            type: The type of noise.  Currently supported types are:
                "gaussian" (draw noise values independently for each
                beam from a Gaussian distribution).
            rate: Noise parameters for angular rates.
            accel: Noise parameters for linear accelerations.
            """

            type: str = field(
                default="gaussian",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            rate: Optional["Sensor.Imu.Noise.Rate"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            accel: Optional["Sensor.Imu.Noise.Accel"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

            @dataclass
            class Rate:
                """
                Noise parameters for angular rates.

                Parameters
                ----------
                mean: For type "gaussian," the mean of the Gaussian
                    distribution from which noise values are drawn.
                stddev: For type "gaussian," the standard deviation of
                    the Gaussian distribution from which noise values
                    are drawn.
                bias_mean: For type "gaussian," the mean of the Gaussian
                    distribution from which bias values are drawn.
                bias_stddev: For type "gaussian," the standard deviation
                    of the Gaussian distribution from which bias values
                    are drawn.
                """

                mean: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                stddev: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                bias_mean: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                bias_stddev: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

            @dataclass
            class Accel:
                """
                Noise parameters for linear accelerations.

                Parameters
                ----------
                mean: For type "gaussian," the mean of the Gaussian
                    distribution from which noise values are drawn.
                stddev: For type "gaussian," the standard deviation of
                    the Gaussian distribution from which noise values
                    are drawn.
                bias_mean: For type "gaussian," the mean of the Gaussian
                    distribution from which bias values are drawn.
                bias_stddev: For type "gaussian," the standard deviation
                    of the Gaussian distribution from which bias values
                    are drawn.
                """

                mean: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                stddev: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                bias_mean: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                bias_stddev: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

    @dataclass
    class Ray:
        """
        These elements are specific to the ray (laser) sensor.

        Parameters
        ----------
        scan:
        range: specifies range properties of each simulated ray
        noise: The properties of the noise model that should be applied
            to generated scans
        """

        scan: Optional["Sensor.Ray.Scan"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        range: Optional["Sensor.Ray.Range"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        noise: Optional["Sensor.Ray.Noise"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

        @dataclass
        class Scan:
            horizontal: Optional["Sensor.Ray.Scan.Horizontal"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            vertical: Optional["Sensor.Ray.Scan.Vertical"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

            @dataclass
            class Horizontal:
                """
                Parameters
                ----------
                samples: The number of simulated rays to generate per
                    complete laser sweep cycle.
                resolution: This number is multiplied by samples to
                    determine the number of range data points returned.
                    If resolution is less than one, range data is
                    interpolated. If resolution is greater than one,
                    range data is averaged.
                min_angle:
                max_angle: Must be greater or equal to min_angle
                """

                samples: int = field(
                    default=640,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                resolution: float = field(
                    default=1.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                min_angle: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                max_angle: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

            @dataclass
            class Vertical:
                """
                Parameters
                ----------
                samples: The number of simulated rays to generate per
                    complete laser sweep cycle.
                resolution: This number is multiplied by samples to
                    determine the number of range data points returned.
                    If resolution is less than one, range data is
                    interpolated. If resolution is greater than one,
                    range data is averaged.
                min_angle:
                max_angle: Must be greater or equal to min_angle
                """

                samples: int = field(
                    default=1,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                resolution: float = field(
                    default=1.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                min_angle: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                max_angle: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

        @dataclass
        class Range:
            """
            specifies range properties of each simulated ray.

            Parameters
            ----------
            min: The minimum distance for each ray.
            max: The maximum distance for each ray.
            resolution: Linear resolution of each ray.
            """

            min: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            max: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            resolution: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Noise:
            """
            The properties of the noise model that should be applied to
            generated scans.

            Parameters
            ----------
            type: The type of noise.  Currently supported types are:
                "gaussian" (draw noise values independently for each
                beam from a Gaussian distribution).
            mean: For type "gaussian," the mean of the Gaussian
                distribution from which noise values are drawn.
            stddev: For type "gaussian," the standard deviation of the
                Gaussian distribution from which noise values are drawn.
            """

            type: str = field(
                default="gaussian",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            mean: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            stddev: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

    @dataclass
    class Sonar:
        """
        These elements are specific to the sonar sensor.

        Parameters
        ----------
        min: Minimum range
        max: Max range
        radius: Radius of the sonar cone at max range.
        """

        min: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        max: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        radius: float = field(
            default=0.5,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Transceiver:
        """
        These elements are specific to a wireless transceiver.

        Parameters
        ----------
        essid: Service set identifier (network name)
        frequency: Specifies the frequency of transmission in MHz
        min_frequency: Only a frequency range is filtered. Here we set
            the lower bound (MHz).
        max_frequency: Only a frequency range is filtered. Here we set
            the upper bound (MHz).
        gain: Specifies the antenna gain in dBi
        power: Specifies the transmission power in dBm
        sensitivity: Mininum received signal power in dBm
        """

        essid: str = field(
            default="wireless",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        frequency: float = field(
            default=2442.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        min_frequency: float = field(
            default=2412.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        max_frequency: float = field(
            default=2484.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        gain: float = field(
            default=2.5,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        power: float = field(
            default=14.5,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        sensitivity: float = field(
            default=-90.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )



from dataclasses import dataclass, field
from typing import Optional

__NAMESPACE__ = "sdformat/v1.0/physics.xsd"


@dataclass
class Physics:
    class Meta:
        name = "physics"

    max_contacts: int = field(
        default=20,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    gravity: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    bullet: Optional["Physics.Bullet"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    ode: Optional["Physics.Ode"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    type: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    update_rate: float = field(
        default=0.0,
        metadata={
            "type": "Attribute",
        },
    )

    @dataclass
    class Bullet:
        dt: float = field(
            default=0.003,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Ode:
        solver: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        constraints: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )



from dataclasses import dataclass, field
from typing import List, Optional
from .geometry import Geometry

__NAMESPACE__ = "sdformat/v1.4/visual.xsd"


@dataclass
class Visual:
    """The visual properties of the link.

    This element specifies the shape of the object (box, cylinder, etc.)
    for visualization purposes.

    Parameters
    ----------
    cast_shadows: If true the visual will cast shadows.
    laser_retro: will be implemented in the future release.
    transparency: The amount of transparency( 0=opaque, 1 = fully
        transparent)
    pose: The reference frame of the visual element, relative to the
        reference frame of the link.
    material: The material of the visual element.
    geometry: The shape of the visual or collision object.
    plugin: A plugin is a dynamically loaded chunk of code. It can exist
        as a child of world, model, and sensor.
    name: Unique name for the visual element within the scope of the
        parent link.
    """

    class Meta:
        name = "visual"

    cast_shadows: bool = field(
        default=True,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    laser_retro: float = field(
        default=0.0,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    transparency: float = field(
        default=0.0,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    pose: str = field(
        default="0 0 0 0 0 0",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
            "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
        },
    )
    material: Optional["Visual.Material"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    geometry: Optional[Geometry] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    plugin: List["Visual.Plugin"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class Material:
        """
        The material of the visual element.

        Parameters
        ----------
        script: Name of material from an installed script file. This
            will override the color element if the script exists.
        shader:
        lighting: If false, dynamic lighting will be disabled
        ambient: The ambient color of a material specified by set of
            four numbers representing red/green/blue, each in the range
            of [0,1].
        diffuse: The diffuse color of a material specified by set of
            four numbers representing red/green/blue/alpha, each in the
            range of [0,1].
        specular: The specular color of a material specified by set of
            four numbers representing red/green/blue/alpha, each in the
            range of [0,1].
        emissive: The emissive color of a material specified by set of
            four numbers representing red/green/blue, each in the range
            of [0,1].
        """

        script: Optional["Visual.Material.Script"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        shader: Optional["Visual.Material.Shader"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        lighting: bool = field(
            default=True,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        ambient: str = field(
            default="0 0 0 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
            },
        )
        diffuse: str = field(
            default="0 0 0 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
            },
        )
        specular: str = field(
            default="0 0 0 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
            },
        )
        emissive: str = field(
            default="0 0 0 1",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){3}\+?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s*",
            },
        )

        @dataclass
        class Script:
            """Name of material from an installed script file.

            This will override the color element if the script exists.

            Parameters
            ----------
            uri: URI of the material script file
            name: Name of the script within the script file
            """

            uri: List[str] = field(
                default_factory=list,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "min_occurs": 1,
                },
            )
            name: str = field(
                default="__default__",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Shader:
            """
            Parameters
            ----------
            normal_map: filename of the normal map
            type: vertex, pixel, normal_map_object_space,
                normal_map_tangent_space
            """

            normal_map: str = field(
                default="__default__",
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            type: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

    @dataclass
    class Plugin:
        """A plugin is a dynamically loaded chunk of code.

        It can exist as a child of world, model, and sensor.

        Parameters
        ----------
        any_element: This is a special element that should not be
            specified in an SDFormat file. It automatically copies child
            elements into the SDFormat element so that a plugin can
            access the data.
        name: A unique name for the plugin, scoped to its parent.
        filename: Name of the shared library to load. If the filename is
            not a full path name, the file will be searched for in the
            configuration paths.
        """

        any_element: List[object] = field(
            default_factory=list,
            metadata={
                "type": "Wildcard",
                "namespace": "##any",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )
        filename: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )



from typing import Any

from .base import ElementBase, Pose


class Origin(ElementBase):
    def __init__(self, *, pose: Pose = None, sdf_version: str) -> None:
        super().__init__(sdf_version=sdf_version)

        if pose is None:
            self.pose = Pose(sdf_version=sdf_version)
        else:
            self.pose = pose

    @classmethod
    def from_specific(cls, specific: Any, *, version: str) -> "Origin":
        return Origin(
            pose=Pose.from_specific(specific.pose, version=version),
            sdf_version=version,
        )



from dataclasses import dataclass, field
from typing import List, Optional
from .sensor import Sensor

__NAMESPACE__ = "sdformat/v1.7/joint.xsd"


@dataclass
class Joint:
    """A joint connects two links with kinematic and dynamic properties.

    By default, the pose of a joint is expressed in the child link
    frame.

    Parameters
    ----------
    parent: Name of the parent link or "world".
    child: Name of the child link. The value "world" may not be
        specified.
    gearbox_ratio: Parameter for gearbox joints.  Given theta_1 and
        theta_2 defined in description for gearbox_reference_body,
        theta_2 = -gearbox_ratio * theta_1.
    gearbox_reference_body: Parameter for gearbox joints.  Gearbox ratio
        is enforced over two joint angles.  First joint angle (theta_1)
        is the angle from the gearbox_reference_body to the parent link
        in the direction of the axis element and the second joint angle
        (theta_2) is the angle from the gearbox_reference_body to the
        child link in the direction of the axis2 element.
    thread_pitch: Parameter for screw joints.
    axis: Parameters related to the axis of rotation for revolute
        joints,       the axis of translation for prismatic joints.
    axis2: Parameters related to the second axis of rotation for
        revolute2 joints and universal joints.
    physics: Parameters that are specific to a certain physics engine.
    pose: A position(x,y,z) and orientation(roll, pitch yaw) with
        respect to the frame named in the relative_to attribute.
    sensor: The sensor tag describes the type and properties of a
        sensor.
    name: A unique name for the joint within the scope of the model.
    type: The type of joint, which must be one of the following:
        (continuous) a hinge joint that rotates on a single axis with a
        continuous range of motion,       (revolute) a hinge joint that
        rotates on a single axis with a fixed range of motion,
        (gearbox) geared revolute joints,       (revolute2) same as two
        revolute joints connected in series,       (prismatic) a sliding
        joint that slides along an axis with a limited range specified
        by upper and lower limits,       (ball) a ball and socket joint,
        (screw) a single degree of freedom joint with coupled sliding
        and rotational motion,       (universal) like a ball joint, but
        constrains one degree of freedom,       (fixed) a joint with
        zero degrees of freedom that rigidly connects two links.
    """

    class Meta:
        name = "joint"

    parent: str = field(
        default="__default__",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    child: str = field(
        default="__default__",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    gearbox_ratio: float = field(
        default=1.0,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    gearbox_reference_body: str = field(
        default="__default__",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    thread_pitch: float = field(
        default=1.0,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    axis: Optional["Joint.Axis"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    axis2: Optional["Joint.Axis2"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    physics: Optional["Joint.Physics"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    pose: Optional["Joint.Pose"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    sensor: List[Sensor] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    type: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class Axis:
        """
        Parameters related to the axis of rotation for revolute joints,
        the axis of translation for prismatic joints.

        Parameters
        ----------
        initial_position: Default joint position for this joint axis.
        xyz: Represents the x,y,z components of the axis unit vector.
            The axis is         expressed in the joint frame unless a
            different frame is expressed in         the expressed_in
            attribute. The vector should be normalized.
        dynamics: An element specifying physical properties of the
            joint. These values are used to specify modeling properties
            of the joint, particularly useful for simulation.
        limit: specifies the limits of this joint
        """

        initial_position: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        xyz: Optional["Joint.Axis.Xyz"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        dynamics: Optional["Joint.Axis.Dynamics"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        limit: Optional["Joint.Axis.Limit"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

        @dataclass
        class Xyz:
            """
            Parameters
            ----------
            value:
            expressed_in: Name of frame in whose coordinates the xyz
                unit vector is expressed.
            """

            value: str = field(
                default="0 0 1",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            expressed_in: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )

        @dataclass
        class Dynamics:
            """An element specifying physical properties of the joint.

            These values are used to specify modeling properties of the
            joint, particularly useful for simulation.

            Parameters
            ----------
            damping: The physical velocity dependent viscous damping
                coefficient of the joint.
            friction: The physical static friction value of the joint.
            spring_reference: The spring reference position for this
                joint axis.
            spring_stiffness: The spring stiffness for this joint axis.
            """

            damping: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            friction: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            spring_reference: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            spring_stiffness: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Limit:
            """
            specifies the limits of this joint.

            Parameters
            ----------
            lower: Specifies the lower joint limit (radians for revolute
                joints, meters for prismatic joints). Omit if joint is
                continuous.
            upper: Specifies the upper joint limit (radians for revolute
                joints, meters for prismatic joints). Omit if joint is
                continuous.
            effort: A value for enforcing the maximum joint effort
                applied. Limit is not enforced if value is negative.
            velocity: A value for enforcing the maximum joint velocity.
            stiffness: Joint stop stiffness.
            dissipation: Joint stop dissipation.
            """

            lower: float = field(
                default=-1e16,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            upper: float = field(
                default=1e16,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            effort: float = field(
                default=-1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            velocity: float = field(
                default=-1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            stiffness: float = field(
                default=100000000.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            dissipation: float = field(
                default=1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

    @dataclass
    class Axis2:
        """
        Parameters related to the second axis of rotation for revolute2 joints
        and universal joints.

        Parameters
        ----------
        initial_position: Default joint position for this joint axis.
        xyz: Represents the x,y,z components of the axis unit vector.
            The axis is         expressed in the joint frame unless a
            different frame is expressed in         the expressed_in
            attribute. The vector should be normalized.
        dynamics: An element specifying physical properties of the
            joint. These values are used to specify modeling properties
            of the joint, particularly useful for simulation.
        limit:
        """

        initial_position: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        xyz: Optional["Joint.Axis2.Xyz"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        dynamics: Optional["Joint.Axis2.Dynamics"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        limit: Optional["Joint.Axis2.Limit"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

        @dataclass
        class Xyz:
            """
            Parameters
            ----------
            value:
            expressed_in: Name of frame in whose coordinates the xyz
                unit vector is expressed.
            """

            value: str = field(
                default="0 0 1",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){2}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            expressed_in: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )

        @dataclass
        class Dynamics:
            """An element specifying physical properties of the joint.

            These values are used to specify modeling properties of the
            joint, particularly useful for simulation.

            Parameters
            ----------
            damping: The physical velocity dependent viscous damping
                coefficient of the joint.  EXPERIMENTAL: if damping
                coefficient is negative and implicit_spring_damper is
                true, adaptive damping is used.
            friction: The physical static friction value of the joint.
            spring_reference: The spring reference position for this
                joint axis.
            spring_stiffness: The spring stiffness for this joint axis.
            """

            damping: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            friction: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            spring_reference: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            spring_stiffness: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Limit:
            """
            Parameters
            ----------
            lower: An attribute specifying the lower joint limit
                (radians for revolute joints, meters for prismatic
                joints). Omit if joint is continuous.
            upper: An attribute specifying the upper joint limit
                (radians for revolute joints, meters for prismatic
                joints). Omit if joint is continuous.
            effort: An attribute for enforcing the maximum joint effort
                applied by Joint::SetForce.  Limit is not enforced if
                value is negative.
            velocity: (not implemented) An attribute for enforcing the
                maximum joint velocity.
            stiffness: Joint stop stiffness. Supported physics engines:
                SimBody.
            dissipation: Joint stop dissipation. Supported physics
                engines: SimBody.
            """

            lower: float = field(
                default=-1e16,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            upper: float = field(
                default=1e16,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            effort: float = field(
                default=-1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            velocity: float = field(
                default=-1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            stiffness: float = field(
                default=100000000.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            dissipation: float = field(
                default=1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

    @dataclass
    class Physics:
        """
        Parameters that are specific to a certain physics engine.

        Parameters
        ----------
        simbody: Simbody specific parameters
        ode: ODE specific parameters
        provide_feedback: If provide feedback is set to true, physics
            engine will compute the constraint forces at this joint.
        """

        simbody: Optional["Joint.Physics.Simbody"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        ode: Optional["Joint.Physics.Ode"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        provide_feedback: bool = field(
            default=False,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

        @dataclass
        class Simbody:
            """
            Simbody specific parameters.

            Parameters
            ----------
            must_be_loop_joint: Force cut in the multibody graph at this
                joint.
            """

            must_be_loop_joint: bool = field(
                default=False,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Ode:
            """
            ODE specific parameters.

            Parameters
            ----------
            cfm_damping: If cfm damping is set to true, ODE will use CFM
                to simulate damping, allows for infinite damping, and
                one additional constraint row (previously used for joint
                limit) is always active.
            implicit_spring_damper: If implicit_spring_damper is set to
                true, ODE will use CFM, ERP to simulate stiffness and
                damping, allows for infinite damping, and one additional
                constraint row (previously used for joint limit) is
                always active.  This replaces cfm_damping parameter in
                SDFormat 1.4.
            fudge_factor: Scale the excess for in a joint motor at joint
                limits. Should be between zero and one.
            cfm: Constraint force mixing for constrained directions
            erp: Error reduction parameter for constrained directions
            bounce: Bounciness of the limits
            max_force: Maximum force or torque used to reach the desired
                velocity.
            velocity: The desired velocity of the joint. Should only be
                set if you want the joint to move on load.
            limit:
            suspension:
            """

            cfm_damping: bool = field(
                default=False,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            implicit_spring_damper: bool = field(
                default=False,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            fudge_factor: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            cfm: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            erp: float = field(
                default=0.2,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            bounce: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            max_force: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            velocity: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            limit: Optional["Joint.Physics.Ode.Limit"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )
            suspension: Optional["Joint.Physics.Ode.Suspension"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

            @dataclass
            class Limit:
                """
                Parameters
                ----------
                cfm: Constraint force mixing parameter used by the joint
                    stop
                erp: Error reduction parameter used by the joint stop
                """

                cfm: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                erp: float = field(
                    default=0.2,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

            @dataclass
            class Suspension:
                """
                Parameters
                ----------
                cfm: Suspension constraint force mixing parameter
                erp: Suspension error reduction parameter
                """

                cfm: float = field(
                    default=0.0,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )
                erp: float = field(
                    default=0.2,
                    metadata={
                        "type": "Element",
                        "namespace": "",
                        "required": True,
                    },
                )

    @dataclass
    class Pose:
        """
        Parameters
        ----------
        value:
        relative_to: Name of frame relative to which the pose is
            applied.
        """

        value: str = field(
            default="0 0 0 0 0 0",
            metadata={
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        relative_to: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
            },
        )



from dataclasses import dataclass, field
from typing import List, Optional
from .collision import Collision
from .sensor import Sensor
from .visual import Visual

__NAMESPACE__ = "sdformat/v1.5/link.xsd"


@dataclass
class Link:
    """A physical link with inertia, collision, and visual properties.

    A link must be a child of a model, and any number of links may exist
    in a model.

    Parameters
    ----------
    gravity: If true, the link is affected by gravity.
    self_collide: If true, the link can collide with other links in the
        model. Two links within a model will collide if
        link1.self_collide OR link2.self_collide. Links connected by a
        joint will never collide.
    kinematic: If true, the link is kinematic only
    must_be_base_link: If true, the link will have 6DOF and be a direct
        child of world.
    velocity_decay: Exponential damping of the link's velocity.
    frame: A frame of reference to which a pose is relative.
    pose: A position(x,y,z) and orientation(roll, pitch yaw) with
        respect to the specified frame.
    inertial: The inertial properties of the link.
    collision: The collision properties of a link. Note that this can be
        different from the visual properties of a link, for example,
        simpler collision models are often used to reduce computation
        time.
    visual: The visual properties of the link. This element specifies
        the shape of the object (box, cylinder, etc.) for visualization
        purposes.
    sensor: The sensor tag describes the type and properties of a
        sensor.
    projector:
    audio_sink: An audio sink.
    audio_source: An audio source.
    battery: Description of a battery.
    name: A unique name for the link within the scope of the model.
    """

    class Meta:
        name = "link"

    gravity: bool = field(
        default=True,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    self_collide: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    kinematic: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    must_be_base_link: bool = field(
        default=False,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    velocity_decay: Optional["Link.VelocityDecay"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    frame: List["Link.Frame"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    pose: Optional["Link.Pose"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    inertial: Optional["Link.Inertial"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    collision: List[Collision] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    visual: List[Visual] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    sensor: List[Sensor] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    projector: Optional["Link.Projector"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    audio_sink: List[str] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    audio_source: List["Link.AudioSource"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    battery: List["Link.Battery"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )

    @dataclass
    class VelocityDecay:
        """
        Exponential damping of the link's velocity.

        Parameters
        ----------
        linear: Linear damping
        angular: Angular damping
        """

        linear: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        angular: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Frame:
        """
        A frame of reference to which a pose is relative.

        Parameters
        ----------
        pose: A position(x,y,z) and orientation(roll, pitch yaw) with
            respect to the specified frame.
        name: Name of the frame. This name must not match another frame
            defined inside the parent that this frame is attached to.
        """

        pose: Optional["Link.Frame.Pose"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

        @dataclass
        class Pose:
            """
            Parameters
            ----------
            value:
            frame: Name of frame which the pose is defined relative to.
            """

            value: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            frame: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )

    @dataclass
    class Pose:
        """
        Parameters
        ----------
        value:
        frame: Name of frame which the pose is defined relative to.
        """

        value: str = field(
            default="0 0 0 0 0 0",
            metadata={
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        frame: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
            },
        )

    @dataclass
    class Inertial:
        """
        The inertial properties of the link.

        Parameters
        ----------
        mass: The mass of the link.
        inertia: The 3x3 rotational inertia matrix. Because the
            rotational inertia matrix is symmetric, only 6 above-
            diagonal elements of this matrix are specified here, using
            the attributes ixx, ixy, ixz, iyy, iyz, izz.
        frame: A frame of reference to which a pose is relative.
        pose: This is the pose of the inertial reference frame, relative
            to the specified reference frame. The origin of the inertial
            reference frame needs to be at the center of gravity. The
            axes of the inertial reference frame do not need to be
            aligned with the principal axes of the inertia.
        """

        mass: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        inertia: Optional["Link.Inertial.Inertia"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        frame: List["Link.Inertial.Frame"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        pose: Optional["Link.Inertial.Pose"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

        @dataclass
        class Inertia:
            """The 3x3 rotational inertia matrix.

            Because the rotational inertia matrix is symmetric, only 6
            above-diagonal elements of this matrix are specified here,
            using the attributes ixx, ixy, ixz, iyy, iyz, izz.
            """

            ixx: float = field(
                default=1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            ixy: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            ixz: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            iyy: float = field(
                default=1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            iyz: float = field(
                default=0.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            izz: float = field(
                default=1.0,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )

        @dataclass
        class Frame:
            """
            A frame of reference to which a pose is relative.

            Parameters
            ----------
            pose: A position(x,y,z) and orientation(roll, pitch yaw)
                with respect to the specified frame.
            name: Name of the frame. This name must not match another
                frame defined inside the parent that this frame is
                attached to.
            """

            pose: Optional["Link.Inertial.Frame.Pose"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

            @dataclass
            class Pose:
                """
                Parameters
                ----------
                value:
                frame: Name of frame which the pose is defined relative
                    to.
                """

                value: str = field(
                    default="0 0 0 0 0 0",
                    metadata={
                        "required": True,
                        "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                    },
                )
                frame: Optional[str] = field(
                    default=None,
                    metadata={
                        "type": "Attribute",
                    },
                )

        @dataclass
        class Pose:
            """
            Parameters
            ----------
            value:
            frame: Name of frame which the pose is defined relative to.
            """

            value: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            frame: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )

    @dataclass
    class Projector:
        """
        Parameters
        ----------
        texture: Texture name
        fov: Field of view
        near_clip: Near clip distance
        far_clip: far clip distance
        frame: A frame of reference to which a pose is relative.
        pose: A position(x,y,z) and orientation(roll, pitch yaw) with
            respect to the specified frame.
        plugin: A plugin is a dynamically loaded chunk of code. It can
            exist as a child of world, model, and sensor.
        name: Name of the projector
        """

        texture: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        fov: float = field(
            default=0.785,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        near_clip: float = field(
            default=0.1,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        far_clip: float = field(
            default=10.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        frame: List["Link.Projector.Frame"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        pose: Optional["Link.Projector.Pose"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        plugin: List["Link.Projector.Plugin"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

        @dataclass
        class Frame:
            """
            A frame of reference to which a pose is relative.

            Parameters
            ----------
            pose: A position(x,y,z) and orientation(roll, pitch yaw)
                with respect to the specified frame.
            name: Name of the frame. This name must not match another
                frame defined inside the parent that this frame is
                attached to.
            """

            pose: Optional["Link.Projector.Frame.Pose"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

            @dataclass
            class Pose:
                """
                Parameters
                ----------
                value:
                frame: Name of frame which the pose is defined relative
                    to.
                """

                value: str = field(
                    default="0 0 0 0 0 0",
                    metadata={
                        "required": True,
                        "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                    },
                )
                frame: Optional[str] = field(
                    default=None,
                    metadata={
                        "type": "Attribute",
                    },
                )

        @dataclass
        class Pose:
            """
            Parameters
            ----------
            value:
            frame: Name of frame which the pose is defined relative to.
            """

            value: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            frame: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )

        @dataclass
        class Plugin:
            """A plugin is a dynamically loaded chunk of code.

            It can exist as a child of world, model, and sensor.

            Parameters
            ----------
            any_element: This is a special element that should not be
                specified in an SDFormat file. It automatically copies
                child elements into the SDFormat element so that a
                plugin can access the data.
            name: A unique name for the plugin, scoped to its parent.
            filename: Name of the shared library to load. If the
                filename is not a full path name, the file will be
                searched for in the configuration paths.
            """

            any_element: List[object] = field(
                default_factory=list,
                metadata={
                    "type": "Wildcard",
                    "namespace": "##any",
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )
            filename: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

    @dataclass
    class AudioSource:
        """
        An audio source.

        Parameters
        ----------
        uri: URI of the audio media.
        pitch: Pitch for the audio media, in Hz
        gain: Gain for the audio media, in dB.
        contact: List of collision objects that will trigger audio
            playback.
        loop: True to make the audio source loop playback.
        frame: A frame of reference to which a pose is relative.
        pose: A position(x,y,z) and orientation(roll, pitch yaw) with
            respect to the specified frame.
        """

        uri: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        pitch: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        gain: float = field(
            default=1.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        contact: Optional["Link.AudioSource.Contact"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        loop: bool = field(
            default=False,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        frame: List["Link.AudioSource.Frame"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        pose: Optional["Link.AudioSource.Pose"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

        @dataclass
        class Contact:
            """
            List of collision objects that will trigger audio playback.

            Parameters
            ----------
            collision: Name of child collision element that will trigger
                audio playback.
            """

            collision: List[str] = field(
                default_factory=list,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "min_occurs": 1,
                },
            )

        @dataclass
        class Frame:
            """
            A frame of reference to which a pose is relative.

            Parameters
            ----------
            pose: A position(x,y,z) and orientation(roll, pitch yaw)
                with respect to the specified frame.
            name: Name of the frame. This name must not match another
                frame defined inside the parent that this frame is
                attached to.
            """

            pose: Optional["Link.AudioSource.Frame.Pose"] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )

            @dataclass
            class Pose:
                """
                Parameters
                ----------
                value:
                frame: Name of frame which the pose is defined relative
                    to.
                """

                value: str = field(
                    default="0 0 0 0 0 0",
                    metadata={
                        "required": True,
                        "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                    },
                )
                frame: Optional[str] = field(
                    default=None,
                    metadata={
                        "type": "Attribute",
                    },
                )

        @dataclass
        class Pose:
            """
            Parameters
            ----------
            value:
            frame: Name of frame which the pose is defined relative to.
            """

            value: str = field(
                default="0 0 0 0 0 0",
                metadata={
                    "required": True,
                    "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
                },
            )
            frame: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                },
            )

    @dataclass
    class Battery:
        """
        Description of a battery.

        Parameters
        ----------
        voltage: Initial voltage in volts.
        name: Unique name for the battery.
        """

        voltage: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )



from dataclasses import dataclass, field
from typing import List, Optional

__NAMESPACE__ = "sdformat/v1.0/sensor.xsd"


@dataclass
class Sensor:
    class Meta:
        name = "sensor"

    origin: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    topic: str = field(
        default="__default",
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    plugin: List["Sensor.Plugin"] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    camera: Optional["Sensor.Camera"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    ray: Optional["Sensor.Ray"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    contact: Optional["Sensor.Contact"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    rfidtag: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    rfid: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    type: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    always_on: bool = field(
        default=False,
        metadata={
            "type": "Attribute",
        },
    )
    update_rate: float = field(
        default=0.0,
        metadata={
            "type": "Attribute",
        },
    )
    visualize: bool = field(
        default=False,
        metadata={
            "type": "Attribute",
        },
    )

    @dataclass
    class Plugin:
        any_element: List[object] = field(
            default_factory=list,
            metadata={
                "type": "Wildcard",
                "namespace": "##any",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )
        filename: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

    @dataclass
    class Camera:
        horizontal_fov: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        image: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        clip: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        save: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        depth_camera: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )

    @dataclass
    class Ray:
        scan: Optional["Sensor.Ray.Scan"] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        range: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

        @dataclass
        class Scan:
            horizontal: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                    "required": True,
                },
            )
            vertical: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Element",
                    "namespace": "",
                },
            )

    @dataclass
    class Contact:
        collision: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        topic: str = field(
            default="__default_topic__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )



import warnings
from typing import List, Any, Dict, Tuple
from itertools import chain

from .base import (
    BoolElement,
    ElementBase,
    FloatElement,
    NamedPoseBearing,
    Pose,
    StringElement,
)
from .frame import Frame
from .origin import Origin
from .plugin import Plugin
from .camera import Camera
from .ray import Ray
from .contact import Contact
from .rfid import RfidTag
from .rfidtag import Rfid
from .imu import Imu
from .forcetorque import ForceTorque
from .gps import Gps
from .sonar import Sonar
from .transceiver import Transceiver
from .altimeter import Altimeter
from .logical_camera import LogicalCamera
from .magnetometer import Magnetometer
from .air_pressure import AirPressure
from .lidar import Lidar
from .navsat import Navsat
from .... import transform as tf


class Sensor(ElementBase):
    """A sensor.

    Parameters
    ----------
    name : str
        A unique name for the sensor. This name must not match another sensor
        under the same parent.
    type : str
        The type name of the sensor. The "ray", "gpu_ray", and "gps" types are
        equivalent to "lidar", "gpu_lidar", and "navsat", respectively. It is
        preferred to use "lidar", "gpu_lidar", and "navsat" since "ray",
        "gpu_ray", and "gps" will be deprecated. The "ray", "gpu_ray", and "gps"
        types are maintained for legacy support. Must be one of:
            - air_pressure
            - altimeter
            - camera
            - contact
            - depth_camera, depth
            - force_torque
            - gps
            - gpu_lidar
            - gpu_ray
            - imu
            - lidar
            - logical_camera
            - magnetometer
            - multicamera
            - navsat
            - ray
            - rfid
            - rfidtag
            - rgbd_camera, rgbd
            - sonar
            - thermal_camera, thermal
            - wireless_receiver
            - wireless_transmitter
    always_on : bool
        If true the sensor will always be updated according to the update rate.
        Default is ``False``.
    update_rate : float
        The frequency at which the sensor data is generated. If set to 0, the
        sensor will generate data every cycle. Default is ``0``.
    visualize : bool
        If true, the sensor is visualized in the GUI. Default is ``False``.
    enable_metrics : bool
        If true, the sensor will publish performance metrics. Default is ``False``.

        .. versionadded:: SDFormat v1.7
    pose : Pose
        The links's initial position (x,y,z) and orientation (roll, pitch, yaw).

        .. versionadded:: SDFormat 1.2
    topic : str
        Name of the topic on which data is published.
    plugins : List[Plugin]
        A list of plugins used to customize the runtime behavior of the
        simulation.
    air_pressure : AirPressure
        Parameters of a Barometer sensor.

        .. versionadded:: SDFormat v1.6
    camera : Camera
        Parameters of a Camera sensor.
    ray : Ray
        Parameters of a Laser sensor.
    contact : Contact
        Parameters of a Contact sensor.
    rfid : Rfid
        Parameters of a RFID sensor.
    rfidtag : RfidTag
        Parameters of a RFID Tag.
    imu : Imu
        Parameters of a Inertial Measurement Unit (IMU).

        .. versionadded:: SDFormat v1.3
    force_torque : ForceTorque
        Parameters of a torque sensor.

        .. versionadded:: SDFormat v1.4
    gps : Gps
        Parameters of a GPS.

        .. versionadded:: SDFormat v1.4
    sonar : Sonar
        parameters of a Sonar.

        .. versionadded:: SDFormat v1.4
    transceiver : Transceiver
        Parameters for a wireless transceiver.

        .. versionadded:: SDFormat v1.4
    altimeter : Altimeter
        Parameters for an Altimeter,

        .. versionadded:: SDFormat v1.5
    lidar : Lidar
        Parameters of a LIDAR sensor.

        .. versionadded:: SDFormat v1.6
    logical_camera : LogicalCamera
        Parameters of a logical camera.
    magnetometer : Magnetometer
        Parameters of a magnetometer.

        .. versionadded:: SDFormat v1.5
    navsat : Navsat
        Parameters of a GPS.

        .. versionadded:: SDFormat v1.7
    sdf_version : str
        The SDFormat version to use when constructing this element.
    origin : Origin
        The link's origin.

        .. deprecated:: SDFormat v1.2
            Use `Sensor.pose` instead.
    frames : List[Frame]
        A list of frames of reference in which poses may be expressed.

        .. deprecated:: SDFormat v1.7
            Use :attr:`Model.frame` instead.
        .. versionadded:: SDFormat v1.5

    Attributes
    ----------
    particle_emitters : List[ParticleEmitter]
        See ``Parameters`` section.
    name : str
        See ``Parameters`` section.
    type : str
        See ``Parameters`` section.
    always_on : bool
        See ``Parameters`` section.
    update_rate : float
        See ``Parameters`` section.
    visualize : bool
        See ``Parameters`` section.
    enable_metrics : bool
        See ``Parameters`` section.
    pose : Pose
        See ``Parameters`` section.
    topic : str
        See ``Parameters`` section.
    plugins : List[Plugin]
        See ``Parameters`` section.
    air_pressure : AirPressure
        See ``Parameters`` section.
    camera : Camera
        See ``Parameters`` section.
    ray : Ray
        See ``Parameters`` section.
    contact : Contact
        See ``Parameters`` section.
    rfid : Rfid
        See ``Parameters`` section.
    rfidtag : RfidTag
        See ``Parameters`` section.
    imu : Imu
        See ``Parameters`` section.
    force_torque : ForceTorque
        See ``Parameters`` section.
    gps : Gps
        See ``Parameters`` section.
    sonar : Sonar
        See ``Parameters`` section.
    transceiver : Transceiver
        See ``Parameters`` section.
    altimeter : Altimeter
        See ``Parameters`` section.
    lidar : Lidar
        See ``Parameters`` section.
    logical_camera : LogicalCamera
        See ``Parameters`` section.
    magnetometer : Magnetometer
        See ``Parameters`` section.
    navsat : Navsat
        See ``Parameters`` section.

    """

    def __init__(
        self,
        *,
        name: str,
        type: str,
        always_on: bool = False,
        update_rate: float = 0,
        visualize: bool = False,
        enable_metrics: bool = False,
        origin: Origin = None,
        pose: Pose = None,
        topic: str = None,
        plugins: List[Plugin],
        air_pressure: AirPressure = None,
        camera: Camera = None,
        ray: Ray = None,
        contact: Contact = None,
        rfid: Rfid = None,
        rfidtag: RfidTag = None,
        imu: Imu = None,
        force_torque: ForceTorque = None,
        gps: Gps = None,
        sonar: Sonar = None,
        transceiver: Transceiver = None,
        frames: List[Frame] = None,
        altimeter: Altimeter = None,
        lidar: Lidar = None,
        logical_camera: LogicalCamera = None,
        magnetometer: Magnetometer = None,
        navsat: Navsat = None,
        sdf_version: str,
    ) -> None:
        super().__init__(sdf_version=sdf_version)

        self.name = name
        self.type = type
        self.always_on = always_on
        self.update_rate = update_rate
        self.visualize = visualize
        self.enable_metrics = enable_metrics
        if origin is None:
            self._origin = Origin(sdf_version=sdf_version)
        elif sdf_version == "1.0":
            self._origin = origin
        else:
            warnings.warn("`origin` is deprecated. Use `Sensor.pose` instead.")
            self._origin = origin
        if sdf_version == "1.0":
            self.pose = self._origin.pose
        elif pose is None:
            self.pose = Pose(sdf_version=sdf_version)
        else:
            self.pose = pose
        self.topic = self.type if topic is None else topic
        self.plugins = [] if plugins is None else plugins
        self.air_pressure = (
            AirPressure(sdf_version=sdf_version)
            if air_pressure is None
            else air_pressure
        )
        self.camera = Camera(sdf_version=sdf_version) if camera is None else camera
        self.ray = Ray(sdf_version=sdf_version) if ray is None else ray
        self.contact = Contact(sdf_version=sdf_version) if contact is None else contact
        self.rfid = Rfid(sdf_version=sdf_version) if rfid is None else rfid
        self.rfidtag = RfidTag(sdf_version=sdf_version) if rfidtag is None else rfidtag
        self.imu = Imu(sdf_version=sdf_version) if imu is None else imu
        self.force_torque = (
            ForceTorque(sdf_version=sdf_version)
            if force_torque is None
            else force_torque
        )
        self.gps = Gps(sdf_version=sdf_version) if gps is None else gps
        self.sonar = Sonar(sdf_version=sdf_version) if sonar is None else sonar
        self.transceiver = (
            Transceiver(sdf_version=sdf_version) if transceiver is None else transceiver
        )
        self._frames = [] if frames is None else frames
        self.altimeter = (
            Altimeter(sdf_version=sdf_version) if altimeter is None else altimeter
        )
        self.lidar = Lidar(sdf_version=sdf_version) if lidar is None else lidar
        self.logical_camera = (
            LogicalCamera(sdf_version=sdf_version)
            if logical_camera is None
            else logical_camera
        )
        self.magnetometer = (
            Magnetometer(sdf_version=sdf_version)
            if magnetometer is None
            else magnetometer
        )
        self.navsat = Navsat(sdf_version=sdf_version) if navsat is None else navsat

        self._origin.pose = self.pose

    @property
    def origin(self):
        warnings.warn(
            "`Sensor.origin` is deprecated since SDFormat v1.2. Use `Sensor.pose` instead."
        )
        return self._origin

    @property
    def frames(self):
        warnings.warn(
            "`Sensor.frames` is deprecated since SDF v1.7."
            " Use `Model.frames` instead and set `Frame.attached_to` to the name of this link.",
            DeprecationWarning,
        )
        return self._frames

    @classmethod
    def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
        sensor_args = {
            "name": specific.name,
            "type": specific.type,
        }
        args_with_default = {
            "always_on": BoolElement,
            "update_rate": FloatElement,
            "visualize": BoolElement,
            "enable_metrics": BoolElement,
            "origin": Origin,
            "pose": Pose,
            "topic": StringElement,
            "air_pressure": AirPressure,
            "camera": Camera,
            "ray": Ray,
            "contact": Contact,
            "rfid": Rfid,
            "rfidtag": RfidTag,
            "imu": Imu,
            "force_torque": ForceTorque,
            "gps": Gps,
            "sonar": Sonar,
            "transceiver": Transceiver,
            "altimeter": Altimeter,
            "lidar": Lidar,
            "logical_camera": LogicalCamera,
            "magnetometer": Magnetometer,
            "navsat": Navsat,
        }
        list_args = {
            "plugin": ("plugins", Plugin),
            "frame": ("frames", Frame),
        }
        standard_args = cls._prepare_standard_args(
            specific, args_with_default, list_args, version=version
        )
        sensor_args.update(standard_args)
        return Sensor(**sensor_args, sdf_version=version)

    def declared_frames(self) -> Dict[str, tf.Frame]:
        declared_frames = {self.name: tf.Frame(3, name=self.name)}

        relevant_config = None
        if self.type == "camera":
            relevant_config = self.camera
        else:
            warnings.warn(f"Sensor type `{self.type}` is not implemented.")

        if relevant_config is not None:
            nested_elements = relevant_config.declared_frames()
            for name, frame in nested_elements.items():
                declared_frames[f"{self.name}::{name}"] = frame

        for frame in self._frames:
            for name, frame in frame.declared_frames().items():
                declared_frames[f"{self.name}::{name}"] = frame

        return declared_frames

    def to_static_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        sensor_frame: str,
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
    ) -> tf.Frame:
        self.pose.to_static_graph(declared_frames, sensor_frame, shape=shape, axis=axis)

        relevant_config = None
        if self.type == "camera":
            relevant_config = self.camera
        else:
            warnings.warn(f"Sensor type `{self.type}` is not implemented.")

        if relevant_config is not None:
            relevant_config.to_static_graph(
                declared_frames,
                sensor_frame,
                seed=seed,
                shape=shape,
                axis=axis,
            )

        for frame in self._frames:
            frame.pose.to_static_graph(
                declared_frames, f"{sensor_frame}::{frame.name}", shape=shape, axis=axis
            )

        return declared_frames[sensor_frame]

    def to_dynamic_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        sensor_frame: str,
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
        apply_state: bool = True,
        _scaffolding: Dict[str, tf.Frame],
    ) -> tf.Frame:
        relevant_config = None
        if self.type == "camera":
            relevant_config = self.camera
        else:
            warnings.warn(f"Sensor type `{self.type}` is not implemented.")

        if relevant_config is not None:
            relevant_config.to_dynamic_graph(
                declared_frames,
                sensor_frame,
                seed=seed,
                shape=shape,
                axis=axis,
                apply_state=apply_state,
                _scaffolding=_scaffolding,
            )

        return declared_frames[sensor_frame]



import numpy as np
from numpy.typing import ArrayLike


def vector_project(a: ArrayLike, b: ArrayLike, *, axis: int = -1) -> np.ndarray:
    """Returns the components of each a along each b.

    Parameters
    ----------
    a : ArrayLike
        A batch of vectors to be projected.
    b : ArrayLike
        A batch of vectors that are being projected onto.
    axis : int
        The data axis of the batches, i.e., along which axis to compute.

    Returns
    -------
    result : ndarray
        A batch of vectors of shape [a.batch_dims, b.batch_dims].


    Notes
    -----
    The function assumes that a and b are broadcastable.

    """

    a = np.asarray(a)
    b = np.asarray(b)

    numerator = np.sum(a * b, axis=axis, keepdims=True)
    denominator = np.sum(b * b, axis=axis, keepdims=True)

    return numerator / denominator * b


def scalar_project(
    a: ArrayLike, b: ArrayLike, *, axis: int = -1, keepdims=False
) -> np.ndarray:
    """Returns the length of the components of each a along each b."""

    projected = vector_project(a, b, axis=axis)
    magnitude = np.linalg.norm(projected, axis=axis, keepdims=keepdims)
    sign = np.sign(np.sum(projected * b, axis=axis, keepdims=keepdims))

    return sign * magnitude


def angle_between(
    vec_a: ArrayLike, vec_b: ArrayLike, *, axis: int = -1, eps=1e-10
) -> np.ndarray:
    """Computes the angle from a to b (in a right-handed frame)

    Notes
    -----
    Implementation is based on this StackOverflow post:
    https://scicomp.stackexchange.com/a/27694
    """

    vec_a = np.asarray(vec_a)[None, :]
    vec_b = np.asarray(vec_b)[None, :]

    if axis >= 0:
        axis += 1

    len_c = np.linalg.norm(vec_a - vec_b, axis=axis)
    len_a = np.linalg.norm(vec_a, axis=axis)
    len_b = np.linalg.norm(vec_b, axis=axis)

    mask = len_a >= len_b
    tmp = np.where(mask, len_a, len_b)
    np.putmask(len_b, ~mask, len_a)
    len_a = tmp

    mask = len_c > len_b
    mu = np.where(mask, len_b - (len_a - len_c), len_c - (len_a - len_b))

    numerator = ((len_a - len_b) + len_c) * mu
    denominator = (len_a + (len_b + len_c)) * ((len_a - len_c) + len_b)

    mask = denominator > eps
    angle = np.divide(numerator, denominator, where=mask)
    np.sqrt(angle, out=angle)
    np.arctan(angle, out=angle)
    angle *= 2
    np.putmask(angle, ~mask, np.pi)
    return angle[0]


# def rotation_direction(vec_a: ArrayLike, vec_b: ArrayLike, *, axis=-1):
#     vec_a = np.asarray(vec_a)[None, :]
#     vec_b = np.asarray(vec_b)[None, :]

#     vec_a = np.moveaxis(vec_a, axis, -1)
#     vec_b = np.moveaxis(vec_b, axis, -1)

#     joint_shape = np.broadcast_shapes(vec_a.shape, vec_b.shape)
#     vec_a = np.broadcast_to(vec_a, joint_shape)
#     vec_b = np.broadcast_to(vec_b, joint_shape)

#     rotation_projection = np.stack((vec_a, vec_b), axis=-2)
#     u, _, v = np.linalg.svd(rotation_projection)
#     directions = np.sign(np.linalg.det(u)*np.linalg.det(v))

#     directions = np.moveaxis(directions, -1, axis)

#     return directions



from dataclasses import dataclass, field
from typing import List, Optional
from .collision import Collision
from .sensor import Sensor
from .visual import Visual

__NAMESPACE__ = "sdformat/v1.0/link.xsd"


@dataclass
class Link:
    class Meta:
        name = "link"

    origin: Optional[str] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    damping: Optional["Link.Damping"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
            "required": True,
        },
    )
    inertial: Optional["Link.Inertial"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    collision: List[Collision] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    visual: List[Visual] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    sensor: List[Sensor] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    projector: Optional["Link.Projector"] = field(
        default=None,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    name: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )
    gravity: bool = field(
        default=True,
        metadata={
            "type": "Attribute",
        },
    )
    self_collide: bool = field(
        default=False,
        metadata={
            "type": "Attribute",
        },
    )
    kinematic: bool = field(
        default=False,
        metadata={
            "type": "Attribute",
        },
    )

    @dataclass
    class Damping:
        linear: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        angular: float = field(
            default=0.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )

    @dataclass
    class Inertial:
        origin: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        inertia: Optional[str] = field(
            default=None,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        mass: float = field(
            default=1.0,
            metadata={
                "type": "Attribute",
            },
        )
        density: float = field(
            default=1.0,
            metadata={
                "type": "Attribute",
            },
        )

    @dataclass
    class Projector:
        texture: str = field(
            default="__default__",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        pose: str = field(
            default="0 0 0 0 0 0",
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
                "pattern": r"(\s*(-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+)\s+){5}((-|\+)?(\d+(\.\d*)?|\.\d+|\d+\.\d+[eE][-\+]?[0-9]+))\s*",
            },
        )
        fov: float = field(
            default=0.785,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        near_clip: float = field(
            default=0.1,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        far_clip: float = field(
            default=10.0,
            metadata={
                "type": "Element",
                "namespace": "",
                "required": True,
            },
        )
        plugin: List["Link.Projector.Plugin"] = field(
            default_factory=list,
            metadata={
                "type": "Element",
                "namespace": "",
            },
        )
        name: Optional[str] = field(
            default=None,
            metadata={
                "type": "Attribute",
                "required": True,
            },
        )

        @dataclass
        class Plugin:
            any_element: List[object] = field(
                default_factory=list,
                metadata={
                    "type": "Wildcard",
                    "namespace": "##any",
                },
            )
            name: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )
            filename: Optional[str] = field(
                default=None,
                metadata={
                    "type": "Attribute",
                    "required": True,
                },
            )



from math import tan
import numpy as np
from numpy.typing import ArrayLike

from ..transform.utils3d import FrustumProjection, EulerRotation
from ..transform.base import Link


class FrustumProjection(FrustumProjection):
    """Frustum based intrinsic camera transformation

    This links behavior is identical to
    :class:`skbot.transform.FrustumProjection`; however, here the camera points
    along the x-axis instead of the z-axis. This is done to match the behavior
    of ignition, which applies the robotic coordinate frame standard instead of
    the computer vision standard here.

    Parameters
    ----------
    hfov : float
        The angle of the viewing frustum in radians. It is assumed to be less than
        pi (180°).
    image_shape : ArrayLike
        The shape (height, width) of the image plane in pixels.

    See Also
    --------
    :class:`skbot.transform.FrustumProjection`

    Notes
    -----
    This function assumes that ``hfov`` is less than pi (180°).

    Points outside the viewing frustum will still be projected; however, their values lie
    outside the applicable pixel range.

    """

    def __init__(self, hfov: float, image_shape: ArrayLike) -> None:
        super().__init__(hfov, image_shape)

        self.initial_rotation = EulerRotation("yz", (-90, 90), degrees=True)

    def transform(self, x: ArrayLike) -> np.ndarray:
        x = self.initial_rotation.transform(x)
        return super().transform(x)



from dataclasses import dataclass, field
from typing import List, Optional
from .actor import Actor
from .light import Light
from .model import Model
from .world import World

__NAMESPACE__ = "sdformat/v1.5/sdf.xsd"


@dataclass
class Sdf:
    """
    SDFormat base element.

    Parameters
    ----------
    world: The world element encapsulates an entire world description
        including: models, scene, physics, joints, and plugins
    model: The model element defines a complete robot or any other
        physical object.
    actor:
    light: The light element describes a light source.
    version: Version number of the SDFormat specification.
    """

    class Meta:
        name = "sdf"

    world: List[World] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    model: List[Model] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    actor: List[Actor] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    light: List[Light] = field(
        default_factory=list,
        metadata={
            "type": "Element",
            "namespace": "",
        },
    )
    version: Optional[str] = field(
        default=None,
        metadata={
            "type": "Attribute",
            "required": True,
        },
    )



import zmq
import subprocess
import socket
import os
import getpass

from . import messages


class Subscriber:
    """Subscribe and listen to Ignition messages.

    Ignition uses ZMQ_ to pass around protocol buffers as means of
    communication. This subscriber enabels python to receive copies of these
    buffers. For more information on the messages published by ignition, and how
    it works, check out the `Ign-Msgs documentation`_ as well as the
    `Ign-Transport documentation`_.


    .. _ZMQ: https://zeromq.org/
    .. _`Ign-Msgs documentation`: https://ignitionrobotics.org/api/msgs/6.4/index.html
    .. _`Ign-Transport documentation`: https://ignitionrobotics.org/api/transport/9.1/index.html
    """

    def __init__(self, topic: str, *, parser=None):
        """Initialize a new subscriber for the given topic.

        Creates an object that uses a context manager to subscribe to
        ign-transport topics and receive messages from it.

        Parameters
        ----------
        topic : str
            The name of the topic to subscribe to as shown by `ign topic -l`.
        parser : function
            A function that deserializes the message. The signature of the parser function
            is ``fn(zmq_message) -> result`` where zmq_message is a 3-tuple of the form
            (zmq_topic, protobuf_message, message_type). If None, the subscriber
            will use a default parser that converts ``protobuf_message`` into a
            ``skbot.ignition.message.<message_type>`` data object.

        Returns
        -------
        self : Subscriber
            A object that can subscribe to ign-transport within a context
            manager

        """
        context = zmq.Context()
        self.socket = context.socket(zmq.SUB)
        self.topic = topic

        # this could be streamlined by speaking the ign-transport discovery protcol
        host_name = socket.gethostname()
        user_name = getpass.getuser()
        self.socket.subscribe(f"@/{host_name}:{user_name}@{topic}")

        if parser is None:
            self.parser = lambda msg: getattr(
                messages, msg[3].decode("utf-8").split(".")[-1]
            )().parse(msg[2])
        else:
            self.parser = parser

    def recv(self, blocking=True, timeout=1000) -> tuple:
        """Receive a message from the topic

        Parameters
        ----------
        blocking : bool
            If True (default) block until a message is received. If False, raise
            zmq.ZMQError if no message is available at the time of query.
        timeout : int
            Time (in ms) to wait for a message to arrive. If the time is
            exceeded, an IOError will be raised. Will wait indefinitely if set
            to `-1`. This only works if ``blocking=True``.

        Returns
        -------
        msg : PyObject
            If a parser was specified during instantiation, returns the result
            of the parser. Otherwise it will use the default parser and return a
            ``skbot.ignition.messages.<message_type>`` data object.
        """

        self.socket.setsockopt(zmq.RCVTIMEO, timeout)

        try:
            if blocking:
                msg = self.socket.recv_multipart()
            else:
                msg = self.socket.recv_multipart(zmq.NOBLOCK)
        except zmq.Again:
            raise IOError(f"Topic {self.topic} did not send a message.")

        result = self.parser(msg)

        return result

    def __enter__(self):
        # weird hack to encourage ign-transport to actually publish camera
        # messages start an echo subscriber and print the messages into the void
        # to make ign realize that something is listening to the topic tracking
        # issue: https://github.com/ignitionrobotics/ign-transport/issues/225
        self.echo_subscriber = subprocess.Popen(
            ["ign", "topic", "-e", "-t", self.topic], stdout=open(os.devnull, "w")
        )

        # this is a bad hack and should be implemented by talking the
        # ign-transport discovery protocol
        result = subprocess.check_output(f"ign topic -i -t {self.topic}", shell=True)
        self.address = (
            result.decode("utf-8")
            .split("\n")[1]
            .split(",")[0]
            .replace("\t", "")
            .replace(" ", "")
        )

        if not self.address:
            self.echo_subscriber.terminate()
            raise IOError(f"Could not identify socket for {self.topic}.")

        self.socket.connect(self.address)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.socket.disconnect(self.address)
        self.echo_subscriber.terminate()



from typing import Dict, Any, Tuple

from .base import ElementBase, Pose, StringElement
from .... import transform as tf


class Frame(ElementBase):
    """A frame of reference.

    Frames are coordinate frames that are rigidly attached to a rigid
    body (or inertial frame) and move together with it. Their primary
    purpose is to make it easier to specify the relative position of
    objects in the simulation.

    .. versionadded:: SDFormat v1.5

    Parameters
    ----------
    name : str
        Name of the frame. It must be unique whithin its scope (model/world),
        i.e., it must not match the name of another frame, link, joint, or model
        within the same scope.
    pose : Pose
        The initial position and orientation of this frame
    attached_to : str
        If specified, this frame is attached to the specified frame. The
        specified frame must be within the same scope and may be defined
        implicitly, i.e., the name of any //frame, //model, //joint, or //link
        within the same scope may be used.

        If missing, this frame is attached to the containing scope's frame.
        Within a //world scope this is the implicit world frame, and within a
        //model scope this is the implicit model frame.

        A frame moves jointly with the frame it is @attached_to. This is
        different from //pose/@relative_to. @attached_to defines how the frame
        is attached to a //link, //model, or //world frame, while
        //pose/@relative_to defines how the frame's pose is represented
        numerically. As a result, following the chain of @attached_to attributes
        must always lead to a //link, //model, //world, or //joint (implicitly
        attached_to its child //link).

        .. versionadded:: SDFormat v1.7
    sdf_version : str
        The SDFormat version to use when constructing this element.

    Attributes
    ----------
    name : str
        See ``Parameters`` section.
    pose : Pose
        See ``Parameters`` section.
    attached_to : str
        See ``Parameters`` section.

    """

    def __init__(
        self,
        *,
        name: str,
        pose: Pose = None,
        attached_to: str = None,
        sdf_version: str,
    ) -> None:
        super().__init__(sdf_version=sdf_version)
        self.name = name
        self.pose = Pose(sdf_version=sdf_version) if pose is None else pose
        self.attached_to = attached_to

        if self.attached_to == "":
            self.attached_to = None

    @classmethod
    def from_specific(cls, specific: Any, *, version: str) -> "ElementBase":
        frame_args = {"name": specific.name}

        args_with_default = {"pose": Pose, "attached_to": StringElement}
        standard_args = cls._prepare_standard_args(
            specific, args_with_default, version=version
        )
        frame_args.update(standard_args)

        return Frame(**frame_args, sdf_version=version)

    def declared_frames(self) -> Dict[str, tf.Frame]:
        return {self.name: tf.Frame(3, name=self.name)}

    def to_static_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
    ) -> tf.Frame:
        parent_name = self.pose.relative_to
        child_name = self.name

        parent = declared_frames[parent_name]
        child = declared_frames[child_name]

        link = self.pose.to_tf_link()
        link(child, parent)

        return declared_frames[self.name]

    def to_dynamic_graph(
        self,
        declared_frames: Dict[str, tf.Frame],
        *,
        seed: int = None,
        shape: Tuple,
        axis: int = -1,
        apply_state: bool = True,
        _scaffolding: Dict[str, tf.Frame],
    ) -> tf.Frame:
        parent_name = self.attached_to
        child_name = self.name

        parent = declared_frames[parent_name]
        child = declared_frames[child_name]

        parent_static = _scaffolding[parent_name]
        child_static = _scaffolding[child_name]

        link = tf.CompundLink(parent_static.links_between(child_static))
        link(parent, child)



from numpy.typing import ArrayLike
from typing import List, Tuple, Union, Callable
import numpy as np
from dataclasses import dataclass, field
from queue import PriorityQueue
import warnings


def DepthFirst(frames: Tuple["Frame"], links: Tuple["Link"]) -> float:
    """Depth-first search metric for Frame.chain_between"""

    return -len(links)


def BreadthFirst(frames: Tuple["Frame"], links: Tuple["Link"]) -> float:
    """Beadth-first search metric for Frame.chain_between"""
    return -1 / len(frames)


@dataclass(order=True)
class QueueItem:
    priority: float
    frames: List["Frame"] = field(compare=False)
    links: List["Link"] = field(compare=False)


class FramePath(str):
    def matches(self, frames: Tuple["Frame"]) -> bool:
        return frames[-1].name == self


class Link:
    """A directional relationship between two Frames

    An abstract class that describes a transformation from a parent frame into a
    child frame. Its default use is to express a vector given in the parent
    frame using the child frame.

    Attributes
    ----------
    transformation : np.ndarray
        A affine matrix describing the transformation from the parent frame to
        the child frame.

    Notes
    -----
    :attr:`Link.transformation` may raise a ``NotImplementedError`` if the link
    doesn't support affine transformation matrices, or if the matrix doesn't exist.

    """

    def __init__(self, parent_dim: int, child_dim: int) -> None:
        self.parent_dim: int = parent_dim
        self.child_dim: int = child_dim

    def __call__(
        self, parent: "Frame", child: "Frame" = None, *, add_inverse: bool = True
    ) -> "Frame":
        """Add this link to the parent frame.

        Parameters
        ----------
        parent : Frame
            The Frame from which vectors originate.
        child : Frame
            The Frame in which vectors are expressed after they were mapped by
            this link's transform. If None, a new child will be created.
        add_inverse : bool
            Also add the inverse link to the child if this Link is invertible.
            Defaults to ``True``.

        Returns
        -------
        child : Frame
            The Frame in which vectors are expressed after they were mapped
            by this link's transform.

        """

        if child is None:
            child = Frame(self.child_dim)

        parent.add_link(self, child)

        if add_inverse:
            try:
                child.add_link(self.invert(), parent)
            except ValueError:
                # inverse does not exist
                pass

        return child

    def invert(self) -> "Frame":
        """Returns a new link that is the inverse of this link.

        The links share parameters, i.e., if the transform of a link changes,
        the transform of its inverse does also.
        """
        return InvertLink(self)

    def transform(self, x: ArrayLike) -> np.ndarray:
        """Expresses the vector x (assumed to be given in the parent's frame) in
        the child's frame.

        Parameters
        ----------
        x : ArrayLike
            The vector expressed in the parent's frame

        Returns
        -------
        y : ArrayLike
            The vector expressed in the child's frame

        """
        raise NotImplementedError

    def __inverse_transform__(self, x: ArrayLike) -> np.ndarray:
        """Transform x (given in the child frame) into the parent frame.

        Parameters
        ----------
        x : ArrayLike
            The vector expressed in the childs's frame

        Returns
        -------
        y : ArrayLike
            The vector expressed in the parents's frame

        """
        raise NotImplementedError


class InvertLink(Link):
    """Inverse of an existing Link.

    This class can be constructed from any link that implements
    __inverse_transform__. It is a tight wrapper around the original link and
    shares any parameters. Accordingly, if the original link updates, so will
    this link.

    Parameters
    ----------
    link : Link
        The link to be inverted. It must implement __inverse_transform__.

    """

    def __init__(self, link: Link) -> None:
        if link.__inverse_transform__.__func__ is Link.__inverse_transform__:
            raise ValueError("Link doesn't implement __inverse_transform__.") from None

        super().__init__(link.child_dim, link.parent_dim)

        self._forward_link = link

    def transform(self, x: ArrayLike) -> np.ndarray:
        return self._forward_link.__inverse_transform__(x)

    def __inverse_transform__(self, x: ArrayLike) -> np.ndarray:
        return self._forward_link.transform(x)

    def __getattr__(self, attr):
        return getattr(self._forward_link, attr)


class Frame:
    """Representation of a coordinate system.

    Each coordinate frame is a node in a directed graph where edges
    (:class:`skbot.transform.Link`) describe transformations between frames. This
    transformation is not limited to neighbours, but works between any two
    frames that share a chain of links pointing from a parent to the
    (grand-)child.

    Parameters
    ----------
    ndim : int
        Number of coordinate dimensions.
    name : str
        The name of this coordinate frame. Defaults to ``None``.

    """

    def __init__(self, ndim: int, *, name: str = None) -> None:
        self._children: List[Tuple(Frame, Link)] = list()
        self.ndim: int = ndim
        self.name = name

    def transform(
        self,
        x: ArrayLike,
        to_frame: Union["Frame", str],
        *,
        ignore_frames: List["Frame"] = None,
    ) -> np.ndarray:
        """Express the vector x in to_frame.

        Express the vector x - assumed to be in this frame - in the coordinate
        frame to_frame. If it is not possible to find a transformation between
        this frame and to_frame a RuntimeError will be raised.

        Parameters
        ----------
        x : ArrayLike
            A vector expressed in this frame.
        to_frame : Frame, str
            The frame in which x should be expressed. If it is a string,
            :func:`~transform` will search for a child frame with the given
            string as name. In case of duplicate names in a frame graph, the
            first one found is used.
        ignore_frames : Frame
            Any frames that should be ignored when searching for a suitable
            transformation chain. Note that this currently does not support
            string aliases.

        Returns
        -------
        x_new : np.ndarray
            The vector x expressed to_frame.

        Raises
        ------
        RuntimeError
            If no suitable chain of transformations can be found, a RuntimeError is raised.

        """

        x_new = np.asarray(x)
        for link in self.links_between(to_frame, ignore_frames=ignore_frames):
            x_new = link.transform(x_new)

        return x_new

    def get_affine_matrix(
        self, to_frame: Union["Frame", str], *, ignore_frames: List["Frame"] = None
    ) -> np.ndarray:
        """Affine transformation matrix to ``to_frame`` (if existant).

        Parameters
        ----------
        to_frame : Frame
            The frame in which x should be expressed.
        ignore_frames : Frame
            Any frames that should be ignored when searching for a suitable
            transformation chain.

        Returns
        -------
        tf_matrix : np.ndarray
            The matrix describing the transformation.

        Raises
        ------
        RuntimeError
            If no suitable chain of transformations can be found, a RuntimeError
            is raised.

        Notes
        -----
        The affine transformation matrix between two frames only exists if the
        transformation chain is linear in ``self.ndim+1`` dimensions. Requesting
        a non-existing affine matrix will raise an Exception. In practice, this
        means that each link along the transformation chain needs to implement
        :attr:`skbot.transform.Link.transformation`.

        """

        tf_matrix = np.eye(self.ndim + 1)
        for link in self.links_between(to_frame, ignore_frames=ignore_frames):
            tf_matrix = link.affine_matrix @ tf_matrix

        return tf_matrix

    def add_link(self, edge: Link, child: "Frame") -> None:
        """Add an edge to the frame graph.

        The edge is directional and points from this frame to another (possibliy
        identical) frame.

        Parameters
        ----------
        edge : Link
            The transformation to add to the graph.
        child : Frame
            The frame that following this link leads to.

        """

        self._children.append((child, edge))

    def _enqueue_children(
        self,
        queue: PriorityQueue,
        queue_item: QueueItem,
        *,
        visited: List["Frame"],
        metric: Callable[[Tuple["Frame"], Tuple[Link]], float],
        max_depth: int,
    ) -> None:
        """Internal logic for :func:chain_between.

        Appends all children that have not been visited previously to
        the search queue.

        """

        frames = queue_item.frames
        sub_chain = queue_item.links

        if len(sub_chain) == max_depth:
            return

        for child, link in self._children:
            if child in visited:
                continue

            new_frames = (*frames, child)
            new_chain = (*sub_chain, link)
            priority = metric(new_frames, new_chain)
            new_item = QueueItem(priority, new_frames, new_chain)
            queue.put(new_item)

    def chain_between(
        self,
        to_frame: Union["Frame", str],
        *,
        ignore_frames: List["Frame"] = None,
        metric: Callable[[Tuple["Frame"], Tuple[Link]], float] = DepthFirst,
        max_depth: int = None,
    ) -> Tuple[List["Frame"], List[Link]]:
        """Get the frames and links between this frame and ``to_frame``.

        .. versionadded:: 0.9.0

        This function searches the frame graph for a chain of transformations
        from the current frame into ``to_frame`` and returns both, the sequence
        of frames and the sequence of links involved in this transformation.

        The sequence of frames will have this frame as its first element,
        and ```to_frame`` as its last element.

        The first element of the returned sequence of links takes as input the
        vector expressed in the current frame; each following element takes as
        input the vector expressed in its predecessor's output frame. The last
        element outputs the vector expressed in ``to_frame``.

        Parameters
        ----------
        to_frame : Frame, str
            The frame in which to express vectors. If ``str``, the graph is
            searched for any node matching that name and the transformation
            chain to the first node matching the name is returned.
        ignore_frames : List[Frame]
            A list of frames to exclude while searching for a transformation
            chain.
        metric : Callable[[Tuple[Frame], Tuple[Link]], float]
            A function to compute the priority of a sub-chain. Sub-chains are
            searched in order of priority starting with the lowest value. The
            first chain that matches ``to_frame`` is returned. You can use a
            custom function that takes a sequence of frames and links involved
            in the sub-chain as input and returns a float (signature
            ``custom_fn(frames, links) -> float``), or you can use a pre-build
            function:

                skbot.transform.metric.DepthFirst
                    The frame graph is searched depth-first with no
                    preference among frames (default).
                skbot.transform.metric.BreadthFirst
                    The frame graph is searched breadth-first with no
                    preference among frames.
        max_depth : int
            If not None, the maximum depth to search, i.e., the maximum length
            of the transform chain.

        Returns
        -------
        frames : List[Frame]
            A list of frames between this frame and ``to_frame`` (inclusive).
        links : List[Link]
            A list of links between this frame and ``to_frame``.

        """
        if max_depth is None:
            max_depth = float("inf")

        if ignore_frames is None:
            ignore_frames = list()

        if isinstance(to_frame, str):
            to_frame = FramePath(to_frame)

        frames = (self,)
        sub_chain = tuple()

        queue = PriorityQueue()
        root_el = QueueItem(metric(frames, sub_chain), frames, sub_chain)
        queue.put(root_el)

        while not queue.empty():
            item: QueueItem = queue.get()
            frames = item.frames
            sub_chain = item.links
            active_frame = frames[-1]

            if isinstance(to_frame, FramePath):
                is_match = to_frame.matches(frames)
            else:
                is_match = active_frame == to_frame

            if is_match:
                break

            ignore_frames.append(active_frame)

            active_frame._enqueue_children(
                queue, item, visited=ignore_frames, metric=metric, max_depth=max_depth
            )

        else:
            raise RuntimeError(
                "Did not find a transformation chain to the target frame."
            )

        return list(frames), sub_chain

    def transform_chain(
        self,
        to_frame: Union["Frame", str],
        *,
        ignore_frames: List["Frame"] = None,
        metric: Callable[[Tuple["Frame"], Tuple[Link]], float] = DepthFirst,
        max_depth: int = None,
    ) -> List[Link]:
        """Get a transformation chain into ``to_frame`` (deprecated).

        .. deprecated:: 0.9.0
            This method is deprecated and will be removed in scikit-bot v1.0.
            Use ``Frame.links_between`` instead.
        .. versionadded:: 0.5.0
            This method was added to Frame.

        This function searches the frame graph for a chain of transformations
        from the current frame into ``to_frame`` and returns the sequence of
        links involved in this transformation. The first element of the returned
        sequence takes as input the vector expressed in the current frame; each
        following element takes as input the vector expressed in its
        predecessor's output frame. The last element outputs the vector
        expressed in ``to_frame``.

        Parameters
        ----------
        to_frame : Frame, str
            The frame in which to express vectors. If ``str``, the graph is
            searched for any node matching that name and the transformation
            chain to the first node matching the name is returned.
        ignore_frames : List[Frame]
            A list of frames to exclude while searching for a transformation
            chain.
        metric : Callable[[Tuple[Frame], Tuple[Link]], float]
            A function to compute the priority of a sub-chain. Sub-chains are
            searched in order of priority starting with the lowest value. The
            first chain that matches ``to_frame`` is returned. You can use a
            custom function that takes a sequence of frames and links involved
            in the sub-chain as input and returns a float (signature
            ``custom_fn(frames, links) -> float``), or you can use a pre-build
            function:

                skbot.transform.metric.DepthFirst
                    The frame graph is searched depth-first with no
                    preference among frames (default).
                skbot.transform.metric.BreadthFirst
                    The frame graph is searched breadth-first with no
                    preference among frames.
        max_depth : int
            If not None, the maximum depth to search, i.e., the maximum length
            of the transform chain.

        """

        warnings.warn(
            "`Frame.transform_chain` is deprecated."
            " Use `Frame.links_between` instead.",
            DeprecationWarning,
        )

        _, links = self.chain_between(
            to_frame, ignore_frames=ignore_frames, metric=metric, max_depth=max_depth
        )

        return links

    def links_between(
        self,
        to_frame: Union["Frame", str],
        *,
        ignore_frames: List["Frame"] = None,
        metric: Callable[[Tuple["Frame"], Tuple[Link]], float] = DepthFirst,
        max_depth: int = None,
    ):
        """Get the links between this frame and ``to_frame``.

        .. versionadded:: 0.9.0

        This function searches the frame graph for a chain of transformations
        from the current frame into ``to_frame`` and returns the sequence of
        links involved in this transformation. The first element of the returned
        sequence takes as input the vector expressed in the current frame; each
        following element takes as input the vector expressed in its
        predecessor's output frame. The last element outputs the vector
        expressed in ``to_frame``.

        Parameters
        ----------
        to_frame : Frame, str
            The frame in which to express vectors. If ``str``, the graph is
            searched for any node matching that name and the transformation
            chain to the first node matching the name is returned.
        ignore_frames : List[Frame]
            A list of frames to exclude while searching for a transformation
            chain.
        metric : Callable[[Tuple[Frame], Tuple[Link]], float]
            A function to compute the priority of a sub-chain. Sub-chains are
            searched in order of priority starting with the lowest value. The
            first chain that matches ``to_frame`` is returned. You can use a
            custom function that takes a sequence of frames and links involved
            in the sub-chain as input and returns a float (signature
            ``custom_fn(frames, links) -> float``), or you can use a pre-build
            function:

                skbot.transform.metric.DepthFirst
                    The frame graph is searched depth-first with no
                    preference among frames (default).
                skbot.transform.metric.BreadthFirst
                    The frame graph is searched breadth-first with no
                    preference among frames.
        max_depth : int
            If not None, the maximum depth to search, i.e., the maximum length
            of the transform chain.

        """

        _, links = self.chain_between(
            to_frame, ignore_frames=ignore_frames, metric=metric, max_depth=max_depth
        )

        return links

    def frames_between(
        self,
        to_frame: Union["Frame", str],
        *,
        include_self: bool = True,
        include_to_frame: bool = True,
        ignore_frames: List["Frame"] = None,
        metric: Callable[[Tuple["Frame"], Tuple[Link]], float] = DepthFirst,
        max_depth: int = None,
    ):
        """Get the frames between this frame and ``to_frame``.

        .. versionadded:: 0.9.0

        This function searches the frame graph for a chain of transformations
        from the current frame into ``to_frame`` and returns the sequence of
        frames between this frame and ``to_frame``.

        Parameters
        ----------
        include_self : bool
            If ``True`` (default) this frame will be added as the first
            element of the returned sequence.
        include_to_frame : bool
            If ``True`` (default) ``to_frame`` will be added as the last
            element of the returned sequence.
        to_frame : Frame, str
            The frame in which to express vectors. If ``str``, the graph is
            searched for any node matching that name and the transformation
            chain to the first node matching the name is returned.
        ignore_frames : List[Frame]
            A list of frames to exclude while searching for a transformation
            chain.
        metric : Callable[[Tuple[Frame], Tuple[Link]], float]
            A function to compute the priority of a sub-chain. Sub-chains are
            searched in order of priority starting with the lowest value. The
            first chain that matches ``to_frame`` is returned. You can use a
            custom function that takes a sequence of frames and links involved
            in the sub-chain as input and returns a float (signature
            ``custom_fn(frames, links) -> float``), or you can use a pre-build
            function:

                skbot.transform.metric.DepthFirst
                    The frame graph is searched depth-first with no
                    preference among frames (default).
                skbot.transform.metric.BreadthFirst
                    The frame graph is searched breadth-first with no
                    preference among frames.
        max_depth : int
            If not None, the maximum depth to search, i.e., the maximum length
            of the transform chain.

        """

        frames, _ = self.chain_between(
            to_frame, ignore_frames=ignore_frames, metric=metric, max_depth=max_depth
        )

        if not include_self:
            frames.pop(0)

        if not include_to_frame:
            frames.pop(-1)

        return frames

    def find_frame(self, path: str, *, ignore_frames: List["Frame"] = None) -> "Frame":
        """Find a frame matching a given path.

        .. versionadded:: 0.3.0
            This method was added to frame.

        This method allows you to find reachable frames using an xpath inspired
        syntax. Path elements are spearated using the `/` character. Each
        element of the path is the name of a frame. For example,
        ``world/link1/link2/gripper``, denotes a sequence of 4 frames with names
        ``["world", "link1", "link2", "gripper"]``. The final frame in the path
        (gripper) is returned.

        By default an element along the path is directly connected to its next
        element. In the previous example this means that there must exist a
        direct link from "world" to "link1". An exception to this rule is the
        use of an ellipsis (...), in which case an element must be connected to
        its next element by a transformation chain (a sequence of
        links).

        The following path elements have special meanings:

            - Ellipsis (``...``)
                Indicates that the previous frame and the next frame are
                connected by a transformation chain instead of being connected
                directly.
            - None (``//``)
                Omitting a name indicates that the name of this frame is None.


        Parameters
        ----------
        xpath : str
            A xpath string describing the frame to search for.
        ignore_frames : List[Frame]
            Any frames that should be ignored when matching the path.

        Returns
        -------
        matched_frame : Frame
            A frame matching the given path.


        Notes
        -----

        In directed graphs there is no clear notion of search order; hence it is
        undefined which frame is found if multiple matches for the path exist.
        In this case an arbitrary match is returned, and you should not count on
        the result to be deterministic.

        Because ``...`` and ``//`` have special meaning, frames with names
        ``"..."`` or ``""`` will be ignored by this method and can not be found.
        Similarly, frames that use slashes, e.g. ``namespace/my_frame``, will be
        ignored and instead the sequences ``["namespace", "my_frame"]`` will be
        matched.

        Each element of the path is assumed to represent a unique frame. This
        means that circular paths will not be matched.

        """

        parts = path.split("/")
        part = parts.pop(0)

        indirect = False
        while len(parts) > 0 and part == "...":
            indirect = True
            part = parts.pop(0)

        if part == "...":
            raise ValueError(f"Path ends with ellipsis: {path}")

        part = None if part == "" else part

        if len(parts) == 0 and self.name == part:
            return self
        elif not indirect and self.name != part:
            raise RuntimeError(f"No match for {path}.")
        elif len(parts) > 0 and self.name == part:
            sub_path = "/".join(parts)
        else:
            parts = ["...", part] + parts
            sub_path = "/".join(parts)

        if ignore_frames is None:
            ignore_frames = []

        local_ignore = [self]
        local_ignore.extend(ignore_frames)

        child_frame: Frame
        for child_frame, _ in self._children:
            if child_frame in local_ignore:
                continue

            try:
                return child_frame.find_frame(sub_path, ignore_frames=local_ignore)
            except RuntimeError:
                continue
        else:
            raise RuntimeError(f"No match for {path}.")


class CustomLink(Link):
    """A link representing a custom transformation.

    Initialize a new link from the callable ``transformation`` that transforms a
    vector from the parent frame to the child frame. This link can represent
    arbitrary transformations between two frames.

    Parameters
    ----------
    parent : Frame
        The frame in which vectors are specified.
    child : Frame
        The frame into which this link transforms vectors.
    transfomration : Callable[[ArrayLike], np.ndarray]
        A callable that takes a vector - in the parent frame - as input and
        returns the vector in the child frame.

    Notes
    -----
    This function does not implement :func:`Link.__inverse_transform__`.

    """

    def __init__(
        self,
        parent_dim: int,
        child_dim: int,
        transformation: Callable[[ArrayLike], np.ndarray],
    ) -> None:
        """Initialize a new custom link."""

        super().__init__(parent_dim, child_dim)
        self._transform = transformation

    def transform(self, x: ArrayLike) -> np.ndarray:
        return self._transform(x)


class CompundLink(Link):
    """A link representing a sequence of other links

    .. versionadded:: 0.4.0

    This link allows you to build a complex link from a sequence of simpler
    links. Upon ``transform`` each link transforms the result of its predecessor
    and the result of the last link is returned. Similarly, when
    __inverse_transform__ is called, each link inverse transformes its
    successor (inverse order).

    Parameters
    ----------
    wrapped_links : List[Link]
        A sequence of link objects. The order is first-to-last, i.e.,
        ``wrapped_links[0]`` is applied first and ``wrapped_links[-1]`` is
        applied last.

    Notes
    -----
    This link will only be invertible if all wrapped links implement
    __inverse_transform__.

    """

    def __init__(self, wrapped_links: List[Link]):
        super().__init__(wrapped_links[0].parent_dim, wrapped_links[-1].child_dim)
        self._links = wrapped_links

    def transform(self, x: ArrayLike) -> np.ndarray:
        for link in self._links:
            x = link.transform(x)

        return x

    def __inverse_transform__(self, x: ArrayLike) -> np.ndarray:
        for link in reversed(self._links):
            x = link.__inverse_transform__(x)

        return x



import numpy as np
from numpy.typing import ArrayLike


def integral(control_points: ArrayLike, t: ArrayLike, *, axis: int = 0) -> float:
    """Estimate the integral along a curve.

    Estimates the integral of a time-parameterized curve along the chosen axis.
    The curve is given by a sequence of control points and their respective
    times. These times need not be spaced out uniformly.

    Parameters
    ----------
    control_points : ArrayLike
        The values of the curve at time points t.
    t : ArrayLike
        The time points at which the curve was evaluated.
    axis : int
        The axis along which to integrate. All other axes are
        treated as batch dimensions.

    Returns
    -------
    estimate : float
        The estimate of the integral under the curve.

    Notes
    -----
    The shapes of control_points and t must match or be broadcastable.

    """

    # This implementation uses the trapezoidal rule

    control_points = np.asarray(control_points)
    t = np.asarray(t)

    f_range = np.arange(control_points.shape[axis] - 1)
    f_lower = np.take(control_points, f_range, axis=axis)
    f_upper = np.take(control_points, f_range + 1, axis=axis)
    f_k = (f_lower + f_upper) / 2

    t_range = np.arange(t.shape[axis] - 1)
    t_lower = np.take(t, t_range, axis=axis)
    t_upper = np.take(t, t_range + 1, axis=axis)
    delta_t = t_upper - t_lower

    return np.sum(f_k * delta_t, axis=axis)


def cumulative_integral(
    control_points: ArrayLike, t: ArrayLike, *, axis: int = 0
) -> np.ndarray:
    """Estimate the cumulative integral along a curve.

    Estimates the cumulative integral of a time-parameterized curve along the
    chosen axis. The curve is given by a sequence of control points and their
    respective times. These times need not be spaced out uniformly.


    Parameters
    ----------
    control_points : ArrayLike
        The values of the curve at time points t.
    t : ArrayLike
        The time points at which the curve was evaluated.
    axis : int
        The axis along which to integrate. All other axes are
        treated as batch dimensions.

    Returns
    -------
    estimate : float
        The estimate of the integral under the curve.

    Notes
    -----
    The shapes of control_points and t must match or be broadcastable.

    """

    # This implementation uses the trapezoidal rule

    control_points = np.asarray(control_points)
    t = np.asarray(t)

    f_k = np.zeros_like(control_points)

    f_range = np.arange(control_points.shape[axis] - 1)
    f_lower = np.take(control_points, f_range, axis=axis)
    f_upper = np.take(control_points, f_range + 1, axis=axis)
    f_k = np.zeros_like(control_points)
    f_k = (f_lower + f_upper) / 2
    f_k = np.insert(f_k, 0, 0, axis=axis)

    t_range = np.arange(t.shape[axis] - 1)
    t_lower = np.take(t, t_range, axis=axis)
    t_upper = np.take(t, t_range + 1, axis=axis)
    delta_t = t_upper - t_lower
    delta_t = np.insert(delta_t, 0, 0, axis=axis)

    return np.cumsum(f_k * delta_t, axis=axis)


__all__ = ["integral", "cumulative_integral"]
